import numpy as np


def closest_level(
    mpp: float | tuple[float, float],
    slide_mpp: float | tuple[float, float],
    downsamples: list[float],
) -> int:
    """Finds the closest downsample level to the given MPP (microns per pixel) value.

    Args:
        mpp: The MPP value(s) to match.
        slide_mpp: The MPP value(s) of the slide.
        downsamples: A list of available downsample levels.

    Returns:
        The index of the closest downsample level.
    """
    scale_factor = np.mean(np.asarray(mpp) / np.asarray(slide_mpp))

    return np.abs(np.asarray(downsamples) - scale_factor).argmin().item()



from pathlib import Path

import numpy as np
from numpy.typing import NDArray

from ratiopath.openslide import OpenSlide


def read_tile(
    slide_path: str | Path,
    tile_coords: tuple[int, int],
    tile_extent: tuple[int, int],
    level: int | None = None,
    resolution: float | tuple[float, float] | None = None,
) -> NDArray:
    assert (level is None) != (resolution is None), (
        "Either level or resolution must be provided"
    )

    with OpenSlide(slide_path) as slide:
        if resolution is not None:
            level = slide.closest_level(resolution)

        slide_region = slide.read_region_relative(tile_coords, level, tile_extent)  # type: ignore[arg-type]

        return np.asarray(slide_region.convert("RGB"))


def tile_overlay(
    overlay_path: str | Path,
    resolution: tuple[int, int],
    roi_coords: tuple[int, int],
    roi_extent: tuple[int, int],
) -> NDArray:
    with OpenSlide(overlay_path) as overlay:
        level = overlay.closest_level(resolution)
        overlay_resolution = overlay.slide_resolution(level)

        resolution_factor = np.asarray(overlay_resolution) / np.asarray(resolution)

        roi_coords = tuple(
            np.round(np.asarray(roi_coords) * resolution_factor).astype(int)
        )
        roi_extent = tuple(
            np.round(np.asarray(roi_extent) * resolution_factor).astype(int)
        )

        overlay_region = overlay.read_region(
            overlay.get_tile_dimensions(roi_coords, level),  # type: ignore[attr-defined]
            level,
            roi_extent,
        )

        return np.asarray(overlay_region.convert("RGB"))


def relative_tile_overlay(
    overlay_path: str | Path,
    resolution: tuple[int, int],
    tile_coords: tuple[int, int],
    relative_roi_coords: tuple[int, int],
    roi_extent: tuple[int, int],
) -> NDArray:
    return tile_overlay(
        overlay_path=overlay_path,
        resolution=resolution,
        roi_coords=tuple(np.asarray(tile_coords) + np.asarray(relative_roi_coords)),
        roi_extent=roi_extent,
    )



from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd
from PIL import Image

from ratiopath.openslide import OpenSlide
from ratiopath.torch.data.openslide_tile_reader import OpenSlideTileReader


class Overlay(OpenSlideTileReader):
    def __init__(
        self,
        overlay_path: str | Path,
        tile_extent: str | int | tuple[int, int],
        slide_resolution: float | tuple[float, float],
        overlay_resolution: float | tuple[float, float] | None = None,
        level: int | str | None = None,
        overlay_mode: str = "1",
        resample_kwargs: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(
            path=overlay_path,
            level=level,
            resolution=overlay_resolution,
            background=None,
        )

        self.slide_resolution = slide_resolution
        self.tile_extent = tile_extent
        self.resample_kwargs = resample_kwargs
        self.overlay_mode = overlay_mode

    def __getitem__(self, coords: tuple[int, int], tile: pd.Series) -> Image.Image:
        level = self._get_from_tile(tile, self.level)
        tile_extent = self._get_from_tile(tile, self.tile_extent)

        with OpenSlide(self.path) as overlay:
            resolution = overlay.slide_resolution(level)

        resolution_factor = np.asarray(resolution) / np.asarray(self.slide_resolution)

        overlay_tile = self.get_openslide_tile(
            tile_coords=tuple(np.round(np.asarray(coords) * resolution_factor)),
            tile_extent=tuple(np.round(np.asarray(tile_extent) * resolution_factor)),
            tile=tile,
        )

        return overlay_tile.convert(self.overlay_mode).resize(
            np.broadcast_to(tile_extent, 2),  # type: ignore[call-overload]
            **(self.resample_kwargs if self.resample_kwargs else {}),
        )



"""Tests for annotation processing functions."""

import io
import json
from math import isclose

import pytest
from shapely import Polygon

from ratiopath.parsers import GeoJSONParser
from ratiopath.tiling.annotations import tile_annotations


class TestMapAnnotations:
    """Test the map_annotations function."""

    @pytest.fixture
    def sample_geojson_content(self):
        """Sample GeoJSON content for testing."""
        return {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "Polygon",
                        "coordinates": [[[0, 0], [10, 0], [10, 10], [0, 10], [0, 0]]],
                    },
                    "properties": {},
                },
                {
                    "type": "Feature",
                    "geometry": {"type": "Point", "coordinates": [5, 5]},
                    "properties": {},
                },
            ],
        }

    def test_map_annotations_basic(self, sample_geojson_content):
        """Test basic functionality of map_annotations."""
        # Create a temporary annotation file
        f = io.StringIO(json.dumps(sample_geojson_content))

        results = tile_annotations(
            annotations=list(GeoJSONParser(f).get_polygons()),
            roi=Polygon([(0, 0), (8, 0), (8, 8), (0, 8)]),
            x=[0, 8, 0, 8],
            y=[0, 0, 8, 8],
            downsample=1,
        )

        for result_polygon, area in zip(results, [64.0, 16.0, 16.0, 4.0], strict=True):
            assert result_polygon.is_valid
            assert isclose(result_polygon.area, area)

    def test_map_annotations_custom_roi(self, sample_geojson_content):
        """Test map_annotations with a custom region of interest."""
        f = io.StringIO(json.dumps(sample_geojson_content))

        results = tile_annotations(
            annotations=list(GeoJSONParser(f).get_polygons()),
            roi=Polygon([(1, 1), (7, 1), (7, 7), (1, 7)]),
            x=[0, 8, 0, 8],
            y=[0, 0, 8, 8],
            downsample=1,
        )

        for result_polygon, area in zip(results, [36.0, 6.0, 6.0, 1.0], strict=True):
            assert result_polygon.is_valid
            assert isclose(result_polygon.area, area)



from functools import partial
from typing import Any

import numpy as np
import pandas as pd
from pandas import DataFrame


def _read_openslide_tiles(path: str, df: DataFrame) -> pd.Series:
    """Read batch of tiles from a whole-slide image using OpenSlide."""
    from PIL import Image

    from ratiopath.openslide import OpenSlide

    with OpenSlide(path) as slide:

        def get_tile(row: pd.Series) -> np.ndarray:
            rgba_region = slide.read_region_relative(
                (row["tile_x"], row["tile_y"]),
                row["level"],
                (row["tile_extent_x"], row["tile_extent_y"]),
            )
            rgb_region = Image.alpha_composite(
                Image.new("RGBA", rgba_region.size, (255, 255, 255)), rgba_region
            ).convert("RGB")
            return np.asarray(rgb_region)

        return df.apply(get_tile, axis=1)


def _read_tifffile_tiles(path: str, df: DataFrame) -> pd.Series:
    """Read batch of tiles from an OME-TIFF file using tifffile."""
    import tifffile
    import zarr

    def get_tile(row: pd.Series, z: zarr.Array) -> np.ndarray:
        tile_slice = z[
            row["tile_y"] : row["tile_y"] + row["tile_extent_y"],
            row["tile_x"] : row["tile_x"] + row["tile_extent_x"],
        ]
        return np.asarray(tile_slice)

    tiles = pd.Series(index=df.index, dtype=object)
    with tifffile.TiffFile(path) as tif:
        for level, group in df.groupby("level"):
            page = tif.series[0].pages[level]
            assert isinstance(page, tifffile.TiffPage)

            z = zarr.open(page.aszarr(), mode="r")
            assert isinstance(z, zarr.Array)

            tiles.loc[group.index] = group.apply(partial(get_tile, z=z), axis=1)

    return tiles


def read_slide_tiles(batch: dict[str, Any]) -> dict[str, Any]:
    """Reads a batch of tiles from a whole-slide image using either OpenSlide or tifffile.

    Args:
        batch:
            - tile_x: X coordinates of tiles relative to the level
            - tile_y: Y coordinates of tiles relative to the level
            - level: Pyramid levels
            - tile_extent_x: Widths of the tiles
            - tile_extent_y: Heights of the tiles

    Returns:
        The input batch with an added `tile` key containing the list of numpy array tiles.
    """
    # Check if it's an OME-TIFF file
    df = pd.DataFrame(batch)
    for path, group in df.groupby("path"):
        if str(path).lower().endswith((".ome.tiff", ".ome.tif")):
            df.loc[group.index, "tile"] = _read_tifffile_tiles(path, group)
        else:
            df.loc[group.index, "tile"] = _read_openslide_tiles(path, group)

    batch["tile"] = df["tile"].tolist()
    return batch



import hashlib
from collections.abc import Callable
from typing import Any


def row_hash(
    row: dict[str, Any],
    column: str = "id",
    algorithm: Callable[[bytes], hashlib._hashlib.HASH] = hashlib.sha256,  # type: ignore[name-defined]
) -> dict[str, Any]:
    """Hashes a row (dictionary) using SHA256 and adds the hash as a new column.

    Args:
        row: The dictionary (row) to hash.
        column: The name of the column to store the hash. Defaults to "id".
        algorithm: The hashing algorithm to use. Defaults to hashlib.sha256.

    Returns:
        The modified row (dictionary) with the SHA256 hash added.
    """
    row[column] = algorithm(str(row).encode()).hexdigest()
    return row
